require 'rest-client'
require 'base64'
require 'json'
require 'websocket-client-simple'

module Trader
  class BitfinexBackend < BaseBackend
    BASE_CUR = Currency.for_code(:BTC)
    QUOTE_CUR = Currency.for_code(:USD)
    MAIN_MARKET = CurrencyPair.new BASE_CUR, QUOTE_CUR

    CUR_EQUIV = {
      'usd' => QUOTE_CUR,
      'btc' => BASE_CUR
    }

    class BalanceAdaptor < RawBalance
      attr_mapped(:amount)
      attr_mapped(:available_amount, 'available')
    end

    class OrderAdaptor < RawAccountOrder
      attr_mapped(:id)
      attr_mapped(:pair) { |r| MAIN_MARKET }
      attr_mapped(:limit?) { |r| r['type'] == 'exchange limit' }
      attr_mapped(:price)
      attr_mapped(:volume, 'original_amount')
      attr_mapped(:executed_volume, 'executed_amount')
      attr_mapped(:instruction) { |r| r['side'] == 'buy' ? Order::BID : Order::ASK }
      attr_mapped(:status) do |r|
        if r['is_live']
          AccountOrder::OPEN
        elsif r['is_cancelled']
          AccountOrder::CANCELED
        else
          AccountOrder::CLOSED
        end
      end
    end

    def initialize
      super :bitfinex
    end

    def get_available_markets
      [MAIN_MARKET]
    end

    def fill_book(_book)
      # TODO: consider book pair

      _book.prepare Time.now

      ob = public_get('book/BTCUSD')
      ob['bids'].each { |o| _book.add_bid(o['price'], o['amount']) }
      ob['asks'].each { |o| _book.add_ask(o['price'], o['amount']) }

      tx = public_get('trades/BTCUSD', { timestamp: one_hour_ago })
      tx.each do |t|
        _book.add_transaction t['price'], t['amount'], Time.at(t['timestamp'])
      end
    end

    def get_session(_cred)
      raise AssertError, 'missing credentials' if !_cred.key?(:api_key) || !_cred.key?(:secret)
      _cred
    end

    def get_balance(_session, _currency)
      raw = auth_post(_session, 'balances')
      wrap_balance raw, _currency
    end

    def withdraw_to_endpoint(_session, _currency, _amount, _endpoint)
      not_supported "usd withdrawals" if _currency == QUOTE_CUR
      execute_btc_withdraw(_session, _amount, _endpoint)
    end

    def get_orders(_session, _pair)
      raw = auth_post(_session, 'orders')
      raw = raw.select { |o| o['type'].include? 'limit' }
      raw.map { |o| wrap_order o }
    end

    def create_order(_session, _pair, _volume, _price, _type)
      params = {
        exchange: "bitfinex",
        symbol: "BTCUSD",
        side: _type == Order::BID ? 'buy' : 'sell',
        amount: _volume.to_s('F')
      }

      if _price.nil?
        params[:price] = _type == Order::BID ? '1.0' : '1000000.0' # some safe values just in case
        params[:type] = 'exchange market'
      else
        params[:price] = _price.to_s('F')
        params[:type] = 'exchange limit'
      end

      wrap_order auth_post(_session, 'order/new', params)
    end

    def fetch_order(_session, _id)
      wrap_order auth_post(_session, 'order/status', { order_id: _id })
    end

    def cancel_order(_session, _id)
      order = wrap_order auth_post(_session, 'order/cancel', { order_id: _id })
      order = fetch_order(_session, _id) while order.status == AccountOrder::OPEN
      order
    end

    def listen_transactions(_pair, &_block)
      WebSocket::Client::Simple.connect('wss://api2.bitfinex.com:3000/ws') do |ws|
        ws.on :open do |event|
          ws.send(JSON.dump({ event: 'subscribe', channel: 'trades', pair: 'BTCUSD' }))
        end

        ws.on :message do |event|
          message = JSON.load(event.data)
          next if message.is_a? Hash

          if message[1] == 'te'
            _block.call(
              pair: _pair,
              ts: Time.at(message[3]),
              direction: message[5] < 0 ? Transaction::SELL : Transaction::BUY,
              price: message[4],
              volume: message[5].abs
            )
          end
        end
      end
    end

    private

    attr_reader :session

    def public_get(_url, _query = {})
      resp = RestClient.get "https://api.bitfinex.com/v1/#{_url}/", params: _query
      JSON.parse resp
    end

    def auth_post(_creds, _url, _data = {})
      payload = build_payload("/v1/#{_url}", _data)
      headers = {
        'Content-Type' => 'application/json',
        'Accept' => 'application/json',
        'X-BFX-PAYLOAD' => payload,
        'X-BFX-SIGNATURE' => sign_payload(_creds, payload),
        'X-BFX-APIKEY' => _creds[:api_key],
      }

      resp = RestClient.post "https://api.bitfinex.com/v1/#{_url}/", nil, headers
      JSON.parse resp
    end

    def build_payload(_url, _params = {})
      payload = {}
      payload['nonce'] = (Time.now.to_f * 10_000).to_i.to_s
      payload['request'] = _url
      payload.merge!(_params) if _params
      Base64.strict_encode64(payload.to_json)
    end

    def sign_payload(_creds, _payload)
      OpenSSL::HMAC.hexdigest('sha384', _creds[:secret], _payload)
    end

    def one_hour_ago
      Time.now.to_i - 3600
    end

    def wrap_balance(_raw, _currency)
      _raw = _raw.find { |b| b["type"] == "exchange" && CUR_EQUIV[b["currency"]] == _currency }
      _raw = { "amount"=> "0.0", "available" => "0.0" } if _raw.nil?
      BalanceAdaptor.new _raw
    end

    def wrap_order(_raw)
      OrderAdaptor.new _raw
    end

    def execute_btc_withdraw(_session, _amount, _endpoint)
      case _endpoint
      when BitcoinAddress
        auth_post(_session, 'withdraw', {
          withdraw_type: 'bitcoin',
          walletselected: 'exchange',
          amount: _amount.to_s('F'),
          address: _endpoint.address
        })
      else
        not_supported "endpoint not supported"
      end
    end

    def process_withdraw(_raw)
      error _raw['message'] if _raw['status'] == 'error'
    end
  end
end