require 'json'
require 'rainbow'
require 'rainbow/ext/string'

module Trader
  # This is needed to prevent one of the subclass files from defining
  # the class without the superclass.
  class GameBackend < BaseBackend; end
end

require 'trade-o-matic/adapters/game_backend/configuration'
require 'trade-o-matic/adapters/game_backend/sfm'
require 'trade-o-matic/adapters/game_backend/state'
require 'trade-o-matic/adapters/game_backend/execute_order'
require 'trade-o-matic/adapters/game_backend/cancel_order'

module Trader
  class GameBackend
    STATUS_MAP = {
      'open' => AccountOrder::OPEN,
      'closed' => AccountOrder::CLOSED,
      'canceled' => AccountOrder::CANCELED
    }

    class GameBalance < RawBalance
      attr_mapped(:amount) { |r| SFM.decode(r['available'] + r['frozen']) }
      attr_mapped(:available_amount) { |r| SFM.decode r['available'] }
    end

    class GameOrder < RawAccountOrder
      attr_mapped(:id)
      attr_mapped(:pair) { |r| CurrencyPair.parse r['pair'] }
      attr_mapped(:price) { |r| SFM.decode r['limit'] }
      attr_mapped(:volume) { |r| SFM.decode r['original_volume'] }
      attr_mapped(:executed_volume) { |r| SFM.decode(r['original_volume'] - r['volume']) }
      attr_mapped(:instruction)
      attr_mapped(:status) { |r| STATUS_MAP[r['status']] }

      def limit?
        true
      end
    end

    attr_reader :verbose, :state

    def initialize
      super :game
      @state = State.new
    end

    def increment_balance(_account_name, _currency, _amount = nil)
      if _amount.nil?
        _amount = _currency.amount
        _currency = _currency.currency
      end

      with_session do
        balance = state.balance_for(_account_name, Currency.for_code(_currency))
        balance['available'] += SFM.encode(_amount)
      end
    end

    def decrement_balance(_account_name, _currency, _amount = nil)
      if _amount.nil?
        _amount = _currency.amount
        _currency = _currency.currency
      end

      with_session do
        balance = state.balance_for(_account_name, Currency.for_code(_currency))
        balance['available'] -= SFM.encode(_amount)
        balance['available'] = 0 if balance['available'] < 0
      end
    end

    # backend implementation:

    def get_available_markets
      with_session(true) do
        state.all_markets
      end
    end

    def fill_book(_book)
      # TODO
    end

    def get_session(_config)
      raise AssertError, 'must provide login information' if _config.nil? or _config[:name].nil?
      _config[:name]
    end

    def get_balance(_session, _currency)
      with_session(true) do
        balance = state.balance_for(_session, _currency)
        GameBalance.new balance
      end
    end

    def get_orders(_session, _pair)
      with_session(true) do
        market = state.market_for(_pair)
        open = market['open'].select { |o| o['account'] = _session }
        closed = market['closed'].select { |o| o['account'] = _session }
        (open + closed).map { |o| GameOrder.new o }
      end
    end

    def fetch_order(_session, _order_id)
      with_session(true) do
        GameOrder.new state.find_order_by_id(_order_id.to_i)[:order]
      end
    end

    def create_order(_session, _pair, _volume, _price, _instruction)
      with_session do
        execute_order = ExecuteOrder.new
        execute_order.state = state
        execute_order.account = _session
        execute_order.pair = _pair
        execute_order.volume = SFM.encode _volume
        execute_order.limit = SFM.encode _price
        execute_order.instruction = _instruction
        execute_order.perform
      end
    end

    def cancel_order(_session, _order_id)
      with_session do
        cancel_order = CancelOrder.new
        cancel_order.state = state
        cancel_order.account = _session
        cancel_order.order_id = _order_id.to_i
        cancel_order.perform
      end
    end

  private

    def with_session(_idemponent = false)
      restore_state_from_file if use_state_file?
      ensure_required_markets
      result = yield
      save_state_to_file if !_idemponent && use_state_file?
      return result
    end

    def use_state_file?
      !config.db_file.nil?
    end

    def restore_state_from_file
      state.load_from_file config.db_file
    end

    def save_state_to_file
      state.store_to_file config.db_file
    end

    def ensure_required_markets
      return false unless config.markets
      config.markets.each { |mc| state.market_for(mc) }
    end

    def config
      Configuration
    end

    def info(_msg, _color = :white)
      puts "GameEx: #{_msg}".color(_color) if verbose
    end
  end
end