require 'rest-client'
require 'json'

module Trader
  class SurbtcBackend < BaseBackend
    API_PREFIX = 'https://www.surbtc.com/api/v1'

    TYPE_MAP = {
      'ask' => Order::ASK,
      'bid' => Order::BID
    }

    STATUS_MAP = {
      'received' => AccountOrder::PENDING,
      'pending' => AccountOrder::OPEN,
      'traded' => AccountOrder::CLOSED,
      'unprepared' => AccountOrder::CLOSED, # error!
      'canceling' => AccountOrder::OPEN,
      'canceled' => AccountOrder::CANCELED
    }

    MAIN_MARKET = CurrencyPair.new(:SATOSHI, :CLP_CENT)
    COP_MARKET = CurrencyPair.new(:SATOSHI, :COP_CENT)

    class SurbtcBalance < RawBalance
      attr_mapped :amount
      attr_mapped :available_amount
    end

    class SurbtcBalanceBTC < RawBalance
      attr_mapped(:amount) { |r| r['amount'] }
      attr_mapped(:available_amount) { |r| r['available_amount'] }
    end

    class SurbtcOrder < RawAccountOrder
      attr_mapped(:id)
      attr_mapped(:pair) { |r| MAIN_MARKET }
      attr_mapped(:price) { |r| r['limit'] }
      attr_mapped(:volume) { |r| r['original_amount'] }
      attr_mapped(:executed_volume) { |r| r['total_exchanged'] }
      attr_mapped(:instruction) { |r| TYPE_MAP[r['type'].downcase] }
      attr_mapped(:status) { |r| STATUS_MAP[r['state']] }
      attr_mapped(:limit?) { |r| r['price_type'] == 'limit' }
    end

    def initialize
      super :surbtc
    end

    def get_available_markets
      [MAIN_MARKET, COP_MARKET]
    end

    def get_balance(_session, _currency)
      res = resource_for "balances/#{currency_code_for(_currency)}", _session
      raw_balance = postprocess res.get

      if _currency == MAIN_MARKET.base
        SurbtcBalanceBTC.new raw_balance['balance']
      else
        SurbtcBalance.new raw_balance['balance']
      end
    end

    def get_orders(_session, _pair)
      res = resource_for "markets/#{market_code_for(_pair)}/orders", _session

      raw_orders = postprocess res.get
      raw_orders['orders'].map { |o| SurbtcOrder.new o }
    end

    def create_order(_session, _pair, _volume, _price, _instruction)
      res = resource_for "markets/#{market_code_for(_pair)}/orders", _session

      raw_order = postprocess res.post({
        order: build_order_json(_price, _volume, _instruction)
      }, :content_type => 'application/json')

      SurbtcOrder.new raw_order['order']
    end

    def fetch_order(_session, _order_id)
      res = resource_for "orders/#{_order_id}", _session

      raw_order = postprocess res.get
      SurbtcOrder.new raw_order['order']
    end

    def cancel_order(_session, _order_id)
      res = resource_for "orders/#{_order_id}", _session

      raw_order = postprocess res.put({
        state: 'canceling'
      }, :content_type => 'application/json')

      while raw_order['order']['state'] == 'canceling'
        raw_order = postprocess res.get
      end

      SurbtcOrder.new raw_order['order']
    end

    private

    def resource_for(_url, _session=nil)
      res = RestClient::Resource.new "#{API_PREFIX}/#{_url}"
      res.options[:headers] = { params: { 'api_key' => _session[:api_key] } } if _session
      res
    end

    def postprocess(_response)
      # TODO: handle error responses
      JSON.parse _response
    end

    def currency_code_for(_currency)
      return 'clp' if _currency.code == :CLP_CENT
      return 'cop' if _currency.code == :COP_CENT
      return 'btc' if _currency.code == :SATOSHI
      _currency.to_s.downcase
    end

    def market_code_for(_pair)
      "#{currency_code_for(_pair.base)}-#{currency_code_for(_pair.quote)}"
    end

    def build_order_json(_price, _volume, _instruction)
      if _price.nil?
        {
          amount: _volume.to_i,
          type: _instruction == Order::BID ? 'bid' : 'ask',
          price_type: 'market'
        }
      else
        {
          limit: _price.to_i,
          amount: _volume.to_i,
          type: _instruction == Order::BID ? 'bid' : 'ask',
          price_type: 'limit'
        }
      end
    end
  end
end